//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//  Evocosm -- An Object-Oriented Framework for Evolutionary Algorithms
//
//  landscape.h
//---------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//
//  This notice applies *only* to this specific expression of this
//  algorithm, and does not imply ownership or invention of the
//  implemented algorithm.
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 2002, 2003 Scott Robert Ladd. All rights reserved.
//  All rights reserved, except as noted herein.
//
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

#if !defined(LIBEVOCOSM_LANDSCAPE_H)
#define LIBEVOCOSM_LANDSCAPE_H

// libevocosm
#include "libevocosm/organism.h"

namespace libevocosm
{
    //! An abstract interface defining a fitness landscape
    /*!
        A "fitness" landscape defines the environment in which an
        organism competes for survival. A class implementing this
        interface will test each organism in a list against some criteria.
        The landscape is tied to the nature of the organism; think of an
        organism as a potential solution to a problem posed by the landscape.
         
        A floating-point organism, for example, could be tested by a fitness
        landscape that represents a function to be maximized. Or, an organsism
        describing the shape of wing could be tested by a landscape
        implementation that understands airflow.
        \param OrganismType - A concrete implementation of the organism template (e.g., the type of organism that "lives" in this landscape)
    */
    template <class OrganismType>
    class landscape : protected globals
    {
    public:
        //! Virtual destructor
        /*!
            A virtual destructor. By default, it does nothing; this is
            a placeholder that identifies this class as a potential base,
            ensuring that objects of a derived class will have their
            destructors called if they are destroyed through a base-class
            pointer.
        */
        virtual ~landscape()
        {
            // nada
        }

        //! Performs fitness testing
        /*!
            Tests a single chromosome for fitness.
            \param a_organism - The organism to be tested by the landscape.
            \param a_show - When true, some implementations of this function will display information about testing
            \return Computed fitness for this organism
        */
        virtual double test(OrganismType & a_organism, bool a_show = false) const = 0;
        
        //! Performs fitness testing
        /*!
            Tests each chromosome in a_population for fitness.
            \param a_population - A vector containing organisms to be tested by the landscape.
            \return A fitness value for the population as a whole; application-defined.
        */
        virtual double test(vector<OrganismType> & a_population) const = 0;
    };
};

#endif
